 /*
 *------------------------------------------------------------------
 *
 * threadimpl.c
 * Nov 98
 * 304-427A
 * Sorin Lerner (based on original implementation from
 * prof. Cooperstock)
 *
 * Implementation of the threads library.
 *
 *------------------------------------------------------------------
 */

#include <stdio.h>
#include <assert.h>
#include <limits.h>
#include "threadimpl.h"
#include "queue.h"

/* 
 * Constant to leave enough space at the top o each stack, so that the
 * OS can save the registers there if it needs to do so.
 */

#define STACK_ADJUST    200 * sizeof(unsigned)

/* 
 * Save all registers on the stack. This must be done by the
 * operating system, since the user process does not have 
 * access to the register sets other than the currently active ones.
 */

#define SAVE_REGS() \
	asm( "ta 3" )


/*
 * The 'in' and 'local' registers of the current register window  
 * are loaded from the stack.
 */

#define RESTORE_REGS() \
	asm( "ld [%sp+0],%l0" ); \
	asm( "ld [%sp+4],%l1" ); \
	asm( "ld [%sp+8],%l2" ); \
	asm( "ld [%sp+12],%l3" ); \
	asm( "ld [%sp+16],%l4" ); \
	asm( "ld [%sp+20],%l5" ); \
	asm( "ld [%sp+24],%l6" ); \
	asm( "ld [%sp+28],%l7" ); \
	asm( "ld [%sp+32],%i0" ); \
	asm( "ld [%sp+36],%i1" ); \
	asm( "ld [%sp+40],%i2" ); \
	asm( "ld [%sp+44],%i3" ); \
	asm( "ld [%sp+48],%i4" ); \
	asm( "ld [%sp+52],%i5" ); \
	asm( "ld [%sp+56],%i6" ); \
	asm( "ld [%sp+60],%i7" )


/*
 * Save the PC and SP of the invoking procedure in the Active thread
 * descriptor.  Note that the PC saved is actually the return address
 * of the callee, which appears here in register %i7, and the SP saved
 * is actually the FP (that is, the SP of the callee prior to the
 * last subroutine call).
 */

#define SAVE_CONTEXT() \
	asm("st %%fp,%0" : "=m"(Active->regs.sp));      /* save sp */ \
	asm("st %%i7,%0" : "=m"(Active->regs.pc))       /* save pc */


/*
 * Loads PC (really the subroutine return address) and SP (of the
 * calling function, presently stored here as the FP) from newly 
 * selected Active thread descriptor. The FP will become the SP 
 * of the invoking procedure on return.
 */

#define RESTORE_CONTEXT() \
	asm("ld %0,%%fp"::"g"(Active->regs.sp));      /* restore sp */ \
	asm("ld %0,%%i7"::"g"(Active->regs.pc))       /* restore pc */


/*
 * global variables
 */

TD TDArray[MAX_THREAD];		/* array of thread descriptors */
TD *Active;			/* the active thread */
Queue *FreeTDs;			/* linked list of free thread descriptors */
Queue *ReadyQueue;		/* linked list of ready threads */
Queue *BlockedQueue;		/* linked list of blocked threads */
int bInitialized = 0;           /* tells us if library has been initialized */


/*
 * Functions to compute the next available id
 */

ThreadId GetNextId(ThreadId id) {
  return id == INT_MAX ? 1 : id+1;
}

int IsInUse(ThreadId id) {
  int i;
  for (i = 0; i < MAX_THREAD; i++)
    if (TDArray[i].inqueue != FreeTDs && TDArray[i].id == id)
      return 1;
  return 0;
}

ThreadId GetNextFreeId() {
  static ThreadId id = 0;

  id = GetNextId(id);
  while (IsInUse(id))
    id = GetNextId(id);
  
  return id;
}

/*
 * Fill in the fields of the TD structure.
 */

void FillTD( TD *td, unsigned pc, void* sp, int priority, ThreadId id) {
  td->id = id;
  td->regs.pc = pc;
  td->regs.sp = (unsigned)sp;
  td->priority = priority;
  td->orig_stack = NULL;
}


/*
 * Prior to removing the active thread from the CPU, it is necessary 
 * to save its context so that it may resume execution from where it
 * left off.  The context switch is performed by placing the Active 
 * thread onto the ReadyQueue and selecting the first element (the
 * thread of highest priority) from the ReadyQueue as the new Active
 * thread.  The context for the new active thread is then restored.   
 * When ContextSwitch returns, execution will resume in the new Active 
 * thread.
 */

void ContextSwitch () {


  SAVE_CONTEXT();

  /*
   * At this point, it is safe to perform a context switch 
   * (change the Active thread) as all pertinent state information 
   * for this thread has been saved on the stack (SAVE_REGS) and 
   * the thread descriptor (SAVE_CONTEXT).
   */
  QueueInsert(ReadyQueue,Active);
  Active = QueueDequeue(ReadyQueue);

  /*
   * We should always get something from the ReadyQueue, since we inserted
   * something right before.
   */
  assert(Active);
  
  /*
   * The subroutine return address has now been set to the newly
   * selected active thread and the FP (which will become the SP 
   * on return) has been set to that of the selected thread.
   */
  RESTORE_CONTEXT();
  
  /*
   * If the active thread has been changed, we will now return 
   * to a different thread.
   */
  return ;
}


/*
 * Initialize the threads library.  The queues are created,
 * all thread descriptors are added to the FreeTDs linked list,
 * and a thread descriptor for the current thread (main) is created.
 * This function must be called before any other library routines are 
 * used.  Otherwise, the library functions should return NOT_INITIALIZED.  
 */

int ThreadsInit() {
  int i;

  if (bInitialized)
    return E_ALREADY_INITIALIZED;

  /* create ReadyQueue, BlockedQueue and FreeTDs */
  
  ReadyQueue = QueueCreate(QUEUE_NOT_OWN_DATA);
  if (NULL == ReadyQueue)
    return E_OUT_OF_MEM;
  
  
  BlockedQueue = QueueCreate(QUEUE_NOT_OWN_DATA);
  if (NULL == BlockedQueue) {
    QueueDestroy(ReadyQueue);
    return E_OUT_OF_MEM;
  }

  FreeTDs = QueueCreate(QUEUE_NOT_OWN_DATA);
  if (NULL == FreeTDs) {
    QueueDestroy(ReadyQueue);
    QueueDestroy(BlockedQueue);
    return E_OUT_OF_MEM;
  }

  /* put all the thread descriptors in the free list */
  for (i = 0; i < MAX_THREAD; i++) {
    FillTD(&TDArray[i],0,0,0,0);
    QueueInsert(FreeTDs,&TDArray[i]);
  }

  /*
   * Allocate and initialize a thread descriptor for the
   * current thread (main), with maximum priority so that it
   * won't be preempted until it yields the process deliberately.
   * Note that the PC and SP will be set later when main() yields.
   * Until then, we don't care about these values.
   */
  Active = QueueDequeue(FreeTDs);
  Active->id = GetNextFreeId();
  Active->priority = MAX_PRIORITY;

  bInitialized = 1;
  
  return OK;
}


/*
 * Creates a new thread by grabbing the next descriptor from
 * the FreeTDs list. TD is initialized and placed on the ReadyQueue.
 * Returns: id of created thread if successful
 *      E_NO_MORE_TD if there are no free TDs.
 *      E_STACK_TOO_SMALL if stack_size is smaller than MIN_STACK_SIZE
 *      E_INVALID_PRI if priority is outside of [MAX_PRIORITY, MIN_PRIORITY]
 */

ThreadId ThreadCreate( unsigned pc, unsigned stack_size, int priority ) {

  void * stack, *orig_stack;
  ThreadId new_id;
  TD* td;

  if (!bInitialized)
    return E_NOT_INITIALIZED;
  
  if (priority > MIN_PRIORITY || priority < MAX_PRIORITY)
    return E_INVALID_PRI;

  if (stack_size < MIN_STACK_SIZE)
    return E_STACK_TOO_SMALL;

  /* 1: get a TD from the freelist */
  td = QueueDequeue(FreeTDs);
  if (!td) 
    return E_NO_MORE_TD;
  
  /* 2: create a stack */
  orig_stack = (void *) malloc( stack_size ) ;
  
  if (orig_stack == NULL)
    return E_OUT_OF_MEM;

  stack = orig_stack + stack_size - STACK_ADJUST;

  /* make sure stack is on a word boundary */
  stack -= (unsigned)stack%4;
  
  /* 3: get the next ThreadId and fill in the fields of the TD. */
  new_id = GetNextFreeId();

  /*
   * NOTE: in code below, use pc-8 as indicated because SPARC
   * processor will add 8 to this saved value when restoring
   * the PC at the return statement. Note also that we not
   * chaging the thread id, since the thread id is valid in the
   * TD structure.
   */
  FillTD( td, pc-8, stack, priority, new_id);
  td->orig_stack = orig_stack; 
	
  /* 4: add TD to ReadyQueue */
  QueueInsert(ReadyQueue,td);

  /* 5: context switch if appropriate? */
  if ( priority < Active->priority ) {
    SAVE_REGS();
    ContextSwitch();
    RESTORE_REGS();
  }

  return new_id;
}


/*
 * Searches for a TD with a given id in an array of TDs. Returns NULL is not
 * found.
 */

TD* FindTDById(TD* array,int size,ThreadId id) {
  int i;
  for (i = 0; i < size; i++)
    if (array[i].inqueue != FreeTDs && array[i].id == id)
      return &array[i];
  return NULL;
}
  
/*
 * Destroys specified thread (or current thread if id == 0)
 * by removing the descriptor from its queue and adding it to the 
 * list of free descriptors. If the active thread is killed, a new 
 * thread is dispatched.  
 * Returns E_INVALID_TID if no thread corresponds to id, otherwise, OK.  
 */

int ThreadDestroy (ThreadId id) {

  TD* td; /* thread descriptor of the thread to destroy */

  if (!bInitialized)
    return E_NOT_INITIALIZED;
  
  /* check if trying to destroy the last ready thread. */
  if (0 == id && QueueIsEmpty(ReadyQueue))
    return E_LAST_READY;
  
  td = (0 == id) ? Active : FindTDById(TDArray,MAX_THREAD,id);

  if (NULL == td)
    return E_INVALID_TID;

  if (td == Active) {        /* destroy ourselves */

    if (td->orig_stack )
      free(td->orig_stack); 

    /* put active thread's process descriptor onto free list */
    QueueEnqueue(FreeTDs,td);

    /* destroy active thread and selects next from ReadyQueue */
    Active = QueueDequeue(ReadyQueue);

    /* now restore context of Active and continue */
    RESTORE_CONTEXT();
  }
  else {                    /* destroy another thread than ourselves */

    /* Remove the TD from its queue. */
    td = QueueDelete(td->inqueue,td);

    /* if this fires, it means we didnd't find the TD! */
    assert(td); 

    if (td->orig_stack )
      free(td->orig_stack);

    /* Add the TD to the free list */
    QueueEnqueue(FreeTDs,td);
  }
  return OK;
}

/*
 * Yield to another thread. This function simply does a context switch.
 */

int ThreadYield() {

  if (!bInitialized)
    return E_NOT_INITIALIZED;
  
  SAVE_REGS();
  ContextSwitch();
  RESTORE_REGS();
  return OK;
}

/*
 * This is exaclty like ContextSwitch, except that instead of inserting
 * the Active thread descriptor onto the ReadyQueue, we add it to the 
 * the end of the BlockedQueue.
 */

void BlockedContextSwitch () {

  SAVE_CONTEXT();

  QueueEnqueue(BlockedQueue,Active);
  Active = QueueDequeue(ReadyQueue);

  /*
   * The way we have our error checking setup, we should never call
   * BlockedContextSwitch if the ReadyQueue is empty. This
   * assert tests this assumption.
   */
  assert(Active);
  
  RESTORE_CONTEXT();

  return ;
}

/*
 * Blocks the calling thread. This is done by adding the Active
 * thread descriptor to the BlockedQueue, and then dispatching the
 * next thread from the ReadyQueue. Returns OK on succes, E_LAST_READY
 * if the last ready thread is being blocked.
 */

int ThreadBlock() {

  if (!bInitialized)
    return E_NOT_INITIALIZED;

  /* Check if this is the last ready thread */
  if (QueueIsEmpty(ReadyQueue))
    return E_LAST_READY;
  
  SAVE_REGS();
  BlockedContextSwitch();
  RESTORE_REGS();
  return OK;
}

/*
 * Wake up the thread with the given id. We do this by removing the thread
 * from the BlockedQueue, and then inserting it into the ReadyQueue.
 * A context switch is done if necessary.
 * On success returns OK. On failure, returns:
 *     E_INVALID_TID if the id is not a valid thread id.
 *     E_NOT_BLOCKED if the thread specified by id is not blocked.
 */

int ThreadWakeUp(ThreadId id) {

  TD* td; /* thread descriptor of the thread to wake up */

  if (!bInitialized)
    return E_NOT_INITIALIZED;
  
  td = FindTDById(TDArray,MAX_THREAD,id);

  if (NULL == td)
    return E_INVALID_TID;

  if (td->inqueue != BlockedQueue)
    return E_NOT_BLOCKED;

  /* Remove the thread descriptor from the blocked queue */
  td = QueueDelete(BlockedQueue,td);
  assert(td);     /* we should find it! */

  /* Insert the thread descriptor in the ready queue */
  QueueInsert(ReadyQueue,td);

  /* Do a context switch if the awoken thread has a higher priority */
  if ( td->priority < Active->priority ) {
    SAVE_REGS();
    ContextSwitch();
    RESTORE_REGS();
  }

  return OK;

}

/* 
 * Change the priority of the given thread. The priority is changed by
 * getting the thread descriptor for the thread, and changing the value of the
 * priority field. If the thread is in the ready queue, then we remove it
 * and then reinsert it, so that it is placed in the apropriate position
 * in the queue.
 *
 * On success returns OK. On failure, returns:
 *      E_INVALID_TID if the id is not a valid thread id.
 *      E_INVALID_PRI if priority is outside of [MAX_PRIORITY, MIN_PRIORITY]
 *
 */

int ThreadChangePriority(ThreadId id,int new_priority) {

  TD* td; /* thread descriptor of the thread to change the priority */

  if (!bInitialized)
    return E_NOT_INITIALIZED;
  
  if (new_priority > MIN_PRIORITY || new_priority < MAX_PRIORITY)
    return E_INVALID_PRI;
  
  td = FindTDById(TDArray,MAX_THREAD,id);

  if (NULL == td)
    return E_INVALID_TID;

  td->priority = new_priority;

  /* if the thread descriptor we've changed was active, we yield */
  if (td == Active)
    ThreadYield();
  else if (td->inqueue == ReadyQueue) {

    /* Delete and reinsert */
    td = QueueDelete(ReadyQueue,td);
    assert(td);     /* we should find it! */
    QueueInsert(ReadyQueue,td);

    if ( td->priority < Active->priority ) {
      SAVE_REGS();
      ContextSwitch();
      RESTORE_REGS();
    }
  }
  return OK;
}

/*
 * This solution does not include the scheduler
 */
int ThreadsUseScheduler() {
  return E_NOT_IMPL;
}

/*
 * A sample (and simple) idle thread.  The code loops forever,
 * printing information about the list of allocated threads.
 * In a real system, Idle would look more like:
 *	void Idle() { while (1) ThreadYield(); }
 */

void Idle() {
  Queue *queue;
  int i;

  while( 1 ) {
    printf( "CPU is idle\n" );
    for( i = 0; i < MAX_THREAD; ++i ) {
      queue = TDArray[i].inqueue;
      if( queue != FreeTDs ) {
        printf( "id = %d, ", TDArray[i].id );
        printf( "pri = %d, ", TDArray[i].priority );
        if( queue == NULL ) printf( "ACTIVE\n");
        else if (queue == ReadyQueue) printf("READY\n");
        else printf( "BLOCKED\n");
      }
    }
    
    (void) sleep( (unsigned) 1 );
    ThreadYield() ;
  }
}
